
import logging
import re
import socket
from functools import partial
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream
from tornado.tcpserver import TCPServer
from tornado.options import define, options, parse_command_line


class Subscription(object):
    def __init__(self, identifier, pattern, stream):
        self.identifier = identifier
        self.pattern = pattern

        self.stream = stream

    def match(self, pattern):
        if self.pattern[-1] == '*':
            return pattern.startswith(self.pattern[:-1])
        else:
            return self.pattern == pattern

    def send(self, message):
        self.stream.write("MESSAGE %s %s\r\n" % (self.identifier, message))


class Connection(object):

    def __init__(self, server, stream):
        self.server = server
        self.stream = stream

        self.stream.read_until('\n', self._on_read_line)
        stream.set_close_callback(self._on_close)

    def _on_read_line(self, data):
        logging.debug("Received raw message: %s" % (data.rstrip('\r\n'),))

        # Callback is only one-time
        self.stream.read_until('\n', self._on_read_line)

        # Some kind of pattern matching for finding what kind of command we receive
        parsed_data = re.match(r"SEND (.+) (.+)\r|\r\n", data)

        if parsed_data:
            pattern, message = parsed_data.groups()

            self.server.send_message_to_clients(pattern, message)
            self.server.send_message_to_relays(pattern, message)

            self.stream.write("OK\r\n")
        else:
            parsed_data = re.match(r"SUBSCRIBE (\d+) (.+)\r|\r\n", data)

            if parsed_data:
                identifier, pattern = parsed_data.groups()

                self.server.add_subscription(Subscription(identifier, pattern, self.stream))

                self.stream.write("OK\r\n")
            else:
                parsed_data = re.match(r"RELAY_MESSAGE (.+) (.+)\r|\r\n", data)

                if parsed_data:
                    pattern, message = parsed_data.groups()

                    self.server.send_message_to_clients(pattern, message)
                else:
                    parsed_data = re.match(r"REGISTER_RELAY", data)

                    if parsed_data:
                        self.server.add_relay(self.stream)
                    else:
                        self.stream.write("ERROR Command is unrecognized\r\n")

    def _on_close(self):
        pass


class ClientConnection(Connection):

    def __init__(self, server, stream):
        super(ClientConnection, self).__init__(server, stream)

        # List of pattern that should be send to this client
        self.subscriptions = []

    def _on_close(self):
        logging.debug("Client has been disconnected")

        for it in self.subscriptions:
            self.server.remove_client(it)

        self.server.remove_relay(self.stream)


class RelayConnection(Connection):

    def __init__(self, server, stream):
        super(RelayConnection, self).__init__(server, stream)

        # Registering self on another server
        self.stream.write("REGISTER_RELAY\r\n")
        self.server.add_relay(stream)

    def _on_close(self):
        logging.debug("Remote relay has been disconnected")

        self.server.remove_relay(self.stream)


class QueueServer(TCPServer):
    def __init__(self):
        super(QueueServer, self).__init__()

        self.subscriptions = []
        self.relays = []

        # Instantiating connections to other servers
        for server in options.servers:
            logging.debug("Connecting to %s server" % (server, ))

            address, port = tuple(server.split(':'))
            port = int(port)

            def _on_connect(stream):
                logging.info("=> Connected")

                RelayConnection(self, stream)

            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            stream = IOStream(s)
            stream.connect((address, port), partial(_on_connect, stream))

    def send_message_to_clients(self, pattern, message):
        logging.debug("\tSending message to clients : %s => %s" % (pattern, message))

        for it in self.subscriptions:
            if it.match(pattern):
                it.send(message)

    def send_message_to_relays(self, pattern, message):
        logging.debug("\tSending message to relays : %s => %s" % (pattern, message))

        for it in self.relays:
            it.write("RELAY_MESSAGE %s %s\r\n" % (pattern, message))

    def add_subscription(self, sub):
        logging.debug("\tCreate new subscription for %s" % (sub.pattern,))
        self.subscriptions.append(sub)

    def remove_client(self, sub):
        self.subscriptions.remove(sub)

    def add_relay(self, relay_stream):
        logging.debug("\t New relay has been registered")
        self.relays.append(relay_stream)

    def remove_relay(self, relay_stream):
        # Some ClientConnection objects is a connection from another server
        if relay_stream in self.relays:
            self.relays.remove(relay_stream)

    def handle_stream(self, stream, address):
        logging.debug("Received incoming TCP stream")
        ClientConnection(self, stream)


if __name__ == "__main__":
    define("port", default=1337, type=int)
    define("servers", multiple=True)
    parse_command_line()

    logging.getLogger().setLevel(logging.DEBUG)

    server = QueueServer()
    server.listen(options.port)

    IOLoop.instance().start()
